import json
import logging
import math
import os
import pickle
import time

import networkx as nx
import osmnx as ox
from django.http import JsonResponse
from django.shortcuts import render

from geder.settings import BASE_DIR

# reprojected_dgm_file = '/Users/gue/Development/geder.at/assets/ogd-10m-at/dhm_at_dgm_10m_2018.tif'  # Update this path to your downloaded DGM file
# cropped_dgm_file = '/Users/gue/Development/geder.at/as#rsets/ogd-10m-at/dhm_at_dgm_10m_2018_sbg_crop.tif'  # Update this path to your downloaded DGM file

COMBINED_ELEVATION_DICT_PATH = os.path.join(BASE_DIR, 'running', 'combined_elevation_dict.pkl')


def get_graph_with_retries(north, south, east, west, network_type='walk', retries=5, delay=5):
    for attempt in range(retries):
        try:
            G = ox.graph_from_bbox(north, south, east, west, network_type=network_type)
            return G
        except json.JSONDecodeError as e:
            logging.error(f"JSON decode error on attempt {attempt + 1}: {e}")
            if attempt < retries - 1:
                time.sleep(delay)
            else:
                raise
        except Exception as e:
            logging.error(f"Attempt {attempt + 1} failed with error: {e}")
            if attempt < retries - 1:
                time.sleep(delay)
            else:
                raise


def planner_view(request):
    bounding_box = {'north': 49.0, 'south': 46.0, 'east': 17.2, 'west': 9.5}
    context = {
        'bounding_box': bounding_box
    }
    return render(request, 'page_running_planner.html', context=context)


def haversine(lat1, lon1, lat2, lon2):
    # Radius of the Earth in kilometers
    R = 6371.0

    # Convert latitude and longitude from degrees to radians
    lat1_rad = math.radians(lat1)
    lon1_rad = math.radians(lon1)
    lat2_rad = math.radians(lat2)
    lon2_rad = math.radians(lon2)

    # Haversine formula
    dlat = lat2_rad - lat1_rad
    dlon = lon2_rad - lon1_rad
    a = math.sin(dlat / 2)**2 + math.cos(lat1_rad) * math.cos(lat2_rad) * math.sin(dlon / 2)**2
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    distance = R * c

    return distance


def load_combined_elevation_dict(path):
    # combined_dict = {}
    if os.path.exists(path):
        with open(path, 'rb') as f:
            print(f"Loading elevation data from {path}")
            return pickle.load(f)
    #             combined_dict.update(elevation_dict)
    # return combined_dict


def add_elevation_to_graph(G, elevation_dict):
    for node in G.nodes:
        elevation = elevation_dict.get(node, 0)  # Default to 0 if no elevation data is found
        G.nodes[node]['elevation'] = elevation
    return G

def calculate_height_gain(G):
    for u, v, data in G.edges(data=True):
        elevation_u = G.nodes[u]['elevation']
        elevation_v = G.nodes[v]['elevation']
        data['height_gain'] = max(0, elevation_v - elevation_u)
    return G

def create_graph(lat, lon, distance):
    G = ox.graph_from_point((lat, lon), dist=distance*1000, network_type='walk')
    return G

def find_route(G, source, target=None):
    if target is None:
        return nx.shortest_path(G, source=source, weight='height_gain', method='dijkstra')
    else:
        return nx.shortest_path(G, source=source, target= target, weight='height_gain', method='dijkstra')

def calculate_route_view(request, source_lat, source_lon, target_lat, target_lon):
    source_lat = float(source_lat)
    source_lon = float(source_lon)
    target_lat = float(target_lat)
    target_lon = float(target_lon)

    distance_km = haversine(source_lat, source_lon, target_lat, target_lon)
    print(f"Calculated distance: {distance_km} km")

    # elevation_dict_paths = [f'elevation_dict_bbox_{i}.pkl' for i in range(len(bounding_boxes))]
    elevation_dict = load_combined_elevation_dict(COMBINED_ELEVATION_DICT_PATH)

    G = create_graph(source_lat, source_lon, distance_km)
    G = add_elevation_to_graph(G, elevation_dict)
    G = calculate_height_gain(G)

    start_node = ox.distance.nearest_nodes(G, source_lon, source_lat)
    target_node = ox.distance.nearest_nodes(G, target_lon, target_lat)
    print(f"Start node: {start_node}, Target node: {target_node}")

    route = find_route(G, start_node, target_node)
    print(f"Route: {route}")

    for node in route:
        print(f"Route node {node}: Elevation {G.nodes[node]['elevation']}")

    route_coords = [(G.nodes[node]['y'], G.nodes[node]['x']) for node in route]
    total_distance = 0.0
    total_elevation_gain = 0.0
    total_elevation_loss = 0.0

    for i in range(len(route) - 1):
        lat1, lon1 = route_coords[i]
        lat2, lon2 = route_coords[i + 1]
        total_distance += haversine(lat1, lon1, lat2, lon2)
        elevation_diff = G.nodes[route[i + 1]]['elevation'] - G.nodes[route[i]]['elevation']
        print(f"Node {route[i]} -> {route[i + 1]}: Elevation {G.nodes[route[i]]['elevation']} -> {G.nodes[route[i + 1]]['elevation']}, Diff: {elevation_diff}")
        if elevation_diff > 0:
            total_elevation_gain += elevation_diff
        else:
            total_elevation_loss += abs(elevation_diff)
    
    print(f"Total Distance: {total_distance} km")
    print(f"Total Elevation Gain: {total_elevation_gain} m")
    print(f"Total Elevation Loss: {total_elevation_loss} m")
    
    return JsonResponse({
        'route': route_coords,
        'total_distance': total_distance,
        'total_elevation_gain': total_elevation_gain,
        'total_elevation_loss': total_elevation_loss,
    })